from __future__ import annotations

import json
import logging
from typing import Any, List, Union

from ouro._resource import SyncAPIResource
from ouro.models import Asset, Comment, Dataset, File, Post, Route, Service

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Assets"]


class Assets(SyncAPIResource):
    def search(
        self,
        query: str = "",
        with_pagination: bool = False,
        **kwargs: Any,
    ) -> Union[List[dict], dict]:
        """
        Search or browse assets.

        When ``query`` is provided, performs hybrid semantic + full-text search.
        When ``query`` is omitted or empty, returns recent assets (browse mode)
        sorted by creation date.  Passing a UUID as the query looks up that
        single asset directly.

        Keyword arguments (all optional):
            asset_type:  "dataset", "post", "file", "service", "route", "quest"
                         (may also be a list, e.g. ["file", "dataset"])
            scope:       "personal" | "org" | "global" | "all"
            org_id:      scope to an organization (UUID)
            team_id:     scope to a team within an org (UUID)
            user_id:     filter by asset owner (UUID)
            visibility:  "public" | "private" | "organization" | "monetized"
            source:      "web" | "api"
            top_level_only: True to exclude child assets
            metadata_filters: dict of metadata key/values, e.g.
                {"file_type": "image", "extension": "csv"}
            limit:  page size (default 20, max 200)
            offset: pagination offset (default 0)

        Returns a list of asset dicts, or a dict with ``data`` and
        ``pagination`` keys when ``with_pagination=True``.
        """
        params: dict[str, Any] = {}
        if query:
            params["query"] = query

        limit = kwargs.pop("limit", 20)
        offset = kwargs.pop("offset", 0)
        params["limit"] = limit
        params["offset"] = offset

        scope = kwargs.pop("scope", None)
        if scope is not None:
            params["scope"] = scope

        metadata_filters = kwargs.pop("metadata_filters", None)
        if metadata_filters is not None:
            params["metadata_filters"] = json.dumps(metadata_filters)

        filter_keys = ("asset_type", "org_id", "team_id", "user_id", "visibility", "source", "top_level_only")
        filters: dict[str, Any] = {}
        for key in filter_keys:
            if key in kwargs:
                filters[key] = kwargs.pop(key)
        if filters:
            params["filters"] = json.dumps(filters)

        params.update(kwargs)
        request = self.client.get("/search/assets", params=params)
        if with_pagination:
            result = self._handle_response(request, with_pagination=True) or {}
            if not isinstance(result, dict):
                return {"data": [], "pagination": None}
            result["data"] = result.get("data") or []
            return result
        return self._handle_response(request) or []

    def retrieve(
        self,
        id: str,
    ) -> Union[Post, Comment, File, Dataset, Service, Route, Asset]:
        """
        Retrieve any asset by its ID, regardless of asset type.
        Automatically determines the asset type and routes to the
        appropriate resource's retrieve method.
        """
        try:
            request = self.client.get(f"/assets/{id}/type")
            data = self._handle_response(request)

            asset_type = data.get("asset_type") if data else None

            if not asset_type:
                raise Exception(f"Asset with id {id} has no asset_type")

            if asset_type == "post":
                return self.ouro.posts.retrieve(id)
            elif asset_type == "comment":
                return self.ouro.comments.retrieve(id)
            elif asset_type == "file":
                return self.ouro.files.retrieve(id)
            elif asset_type == "dataset":
                return self.ouro.datasets.retrieve(id)
            elif asset_type == "service":
                return self.ouro.services.retrieve(id)
            elif asset_type == "route":
                return self.ouro.routes.retrieve(id)
            else:
                log.warning(
                    f"Unknown asset type: {asset_type}. Cannot retrieve full asset details via API."
                )
                raise Exception(
                    f"Asset type '{asset_type}' is not supported by the unified retrieve method. "
                    f"Please use the specific resource's retrieve method if available."
                )
        except Exception as e:
            error_str = str(e).lower()
            if "not found" in error_str or "404" in error_str or "no rows" in error_str:
                raise Exception(f"Asset with id {id} not found") from e
            raise



from __future__ import annotations

from typing import Optional

import httpx
from typing_extensions import Literal

from ._utils import is_dict

__all__ = [
    "BadRequestError",
    "AuthenticationError",
    "PermissionDeniedError",
    "NotFoundError",
    "ConflictError",
    "UnprocessableEntityError",
    "RateLimitError",
    "InternalServerError",
]


class OuroError(Exception):
    pass


class APIError(OuroError):
    message: str
    request: httpx.Request

    body: object | None
    """The API response body.

    If the API responded with a valid JSON structure then this property will be the
    decoded result.

    If it isn't a valid JSON structure then this will be the raw response.

    If there was no response associated with this error then it will be `None`.
    """

    code: Optional[str] = None
    param: Optional[str] = None
    type: Optional[str]

    def __init__(
        self, message: str, request: httpx.Request, *, body: object | None
    ) -> None:
        super().__init__(message)
        self.request = request
        self.message = message
        self.body = body

        if is_dict(body):
            self.code = body.get("code")
            self.param = body.get("param")
            self.type = body.get("type")
        else:
            self.code = None
            self.param = None
            self.type = None


class APIResponseValidationError(APIError):
    response: httpx.Response
    status_code: int

    def __init__(
        self,
        response: httpx.Response,
        body: object | None,
        *,
        message: str | None = None,
    ) -> None:
        super().__init__(
            message or "Data returned by API invalid for expected schema.",
            response.request,
            body=body,
        )
        self.response = response
        self.status_code = response.status_code


class APIStatusError(APIError):
    """Raised when an API response has a status code of 4xx or 5xx."""

    response: httpx.Response
    status_code: int
    request_id: str | None

    def __init__(
        self, message: str, *, response: httpx.Response, body: object | None
    ) -> None:
        super().__init__(message, response.request, body=body)
        self.response = response
        self.status_code = response.status_code
        self.request_id = response.headers.get("x-request-id")


class APIConnectionError(APIError):
    def __init__(
        self, *, message: str = "Connection error.", request: httpx.Request
    ) -> None:
        super().__init__(message, request, body=None)


class APITimeoutError(APIConnectionError):
    def __init__(self, request: httpx.Request) -> None:
        super().__init__(message="Request timed out.", request=request)


class BadRequestError(APIStatusError):
    status_code: Literal[400] = (
        400  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class AuthenticationError(APIStatusError):
    status_code: Literal[401] = (
        401  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class PermissionDeniedError(APIStatusError):
    status_code: Literal[403] = (
        403  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class NotFoundError(APIStatusError):
    status_code: Literal[404] = (
        404  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class ConflictError(APIStatusError):
    status_code: Literal[409] = (
        409  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class UnprocessableEntityError(APIStatusError):
    status_code: Literal[422] = (
        422  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class RateLimitError(APIStatusError):
    status_code: Literal[429] = (
        429  # pyright: ignore[reportIncompatibleVariableOverride]
    )


class InternalServerError(APIStatusError):
    pass



from uuid import UUID

from .content import description_to_markdown, tiptap_to_markdown
from .openapi import get_custom_openapi, ouro_field

__all__ = [
    "ouro_field",
    "get_custom_openapi",
    "is_valid_uuid",
    "tiptap_to_markdown",
    "description_to_markdown",
]


def is_valid_uuid(uuid_string: str) -> bool:
    try:
        # Attempt to create a UUID object
        uuid_obj = UUID(uuid_string)
        return str(uuid_obj) == uuid_string
    except ValueError:
        return False



# Utility functions for using Plotly html files in Ouro


from __future__ import annotations

import os
from typing import Optional, Tuple

MODEBAR_CSS = """
<style>
/* Move Plotly toolbar to the top-left */
.plotly-graph-div .modebar {
  left: 0 !important;
  right: auto !important;
}
</style>
"""


THEME_JS = """
<script>
(function(){
  function applyTheme(theme){
    try {
      var dark = (theme === 'dark');
      var update = {
        template: dark ? 'plotly_dark' : 'plotly_white',
        paper_bgcolor: dark ? '#262626' : '#f8f8f8',
        plot_bgcolor: dark ? '#262626' : '#f8f8f8',
        font: {color: dark ? '#e5e5e5' : '#262626'},
        xaxis: {gridcolor: dark ? '#333333' : '#dddddd', zerolinecolor: dark ? '#666666' : '#999999'},
        yaxis: {gridcolor: dark ? '#333333' : '#dddddd', zerolinecolor: dark ? '#666666' : '#999999'}
      };
      var gd = document.getElementById('plot') || document.querySelector('div.plotly-graph-div');
      if (gd && window.Plotly && window.Plotly.relayout) {
        window.Plotly.relayout(gd, update);
      }
      document.documentElement.style.background = update.paper_bgcolor;
      document.body.style.background = update.paper_bgcolor;
    } catch (e) {}
  }
  window.addEventListener('message', function(event){
    var data = event && event.data ? event.data : {};
    if (data && data.type === 'ouro-theme') {
      applyTheme(data.theme);
    }
  }, false);
  try {
    var prefersDark = window.matchMedia && window.matchMedia('(prefers-color-scheme: dark)').matches;
    applyTheme(prefersDark ? 'dark' : 'light');
  } catch (e) {}
})();
</script>
"""


def _normalize_dir_url(url: str) -> str:
    return url.rstrip("/")


def _plotly_dir_from_base(base_url: str) -> str:
    base = _normalize_dir_url(base_url)
    # If the base already ends with /plotly, don't append again
    if base.endswith("/plotly"):
        return base
    return f"{base}/plotly"


def _resolve_assets_dir_url(override_dir_url: Optional[str] = None) -> Optional[str]:
    """Resolve directory URL that contains plotly-modebar.css and plotly-theme.js.

    Precedence:
    1) override_dir_url (used as-is)
    2) OURO_PLOTLY_ASSETS_URL (used as-is)
    3) OURO_PUBLIC_ASSETS_URL (append /plotly)
    4) OURO_FRONTEND_URL (append /plotly)
    """
    if override_dir_url:
        return _normalize_dir_url(override_dir_url)

    exact = os.getenv("OURO_PLOTLY_ASSETS_URL")
    if exact:
        return _normalize_dir_url(exact)

    app_base = os.getenv("OURO_PUBLIC_ASSETS_URL") or os.getenv("OURO_FRONTEND_URL")
    if app_base:
        return _plotly_dir_from_base(app_base)

    return None


def build_plotly_asset_tags(
    assets_dir_url: Optional[str] = None,
) -> Tuple[str, str]:
    """Return (css_tag, script_tag) for Plotly UI/theme.

    If a directory URL is available, return external link/script tags that point to
    the app-hosted static assets. Otherwise, fall back to inline CSS/JS.
    """
    resolved = _resolve_assets_dir_url(assets_dir_url)
    if resolved:
        css = f'<link rel="stylesheet" href="{resolved}/plotly-modebar.css" />'
        js = f'<script src="{resolved}/plotly-theme.js"></script>'
        return css, js
    return MODEBAR_CSS, THEME_JS


def inject_assets_into_html(
    html: str,
    assets_dir_url: Optional[str] = None,
) -> str:
    """Inject CSS/JS tags into a Plotly HTML document.

    Prefers placing CSS before </head> and JS before </body>. Falls back to
    appending if the markers are missing.
    """
    css_tag, js_tag = build_plotly_asset_tags(assets_dir_url)

    # Insert CSS in <head> when possible
    if "</head>" in html:
        html = html.replace("</head>", f"{css_tag}\n</head>")
    else:
        html += css_tag

    # Insert JS before </body> when possible
    if "</body>" in html:
        html = html.replace("</body>", f"{js_tag}\n</body>")
    else:
        html += js_tag

    return html



from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, Union

from ouro._resource import SyncAPIResource, _coerce_description, _strip_none
from ouro.models import Post

from .content import Content, Editor

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Posts"]


class Posts(SyncAPIResource):
    def _resolve_content(
        self,
        content: Optional["Content"],
        content_markdown: Optional[str],
        content_path: Optional[str],
    ) -> "Content":
        provided = [
            ("content", content is not None),
            ("content_markdown", content_markdown is not None),
            ("content_path", content_path is not None),
        ]
        selected = [name for name, is_set in provided if is_set]
        if len(selected) > 1:
            raise ValueError(
                f"Provide only one of content, content_markdown, or content_path (got: {', '.join(selected)})."
            )

        if content is not None:
            return content

        markdown: Optional[str] = content_markdown
        if content_path is not None:
            path = Path(content_path).expanduser()
            if not path.exists():
                raise ValueError(f"content_path not found: {content_path}")
            if not path.is_file():
                raise ValueError(f"content_path must point to a file: {content_path}")
            if path.suffix.lower() not in {".md", ".markdown"}:
                raise ValueError("content_path must be a .md or .markdown file.")
            markdown = path.read_text(encoding="utf-8")

        if markdown is None:
            raise ValueError(
                "No post body provided. Pass one of: content, content_markdown, or content_path."
            )

        resolved_content = self.Content()
        resolved_content.from_markdown(markdown)
        return resolved_content

    def Editor(self, **kwargs) -> Editor:
        """Create an Editor instance connected to the Ouro client."""
        return Editor(_ouro=self.ouro, **kwargs)

    def Content(self, **kwargs) -> "Content":
        """Create a Content instance connected to the Ouro client."""
        return Content(_ouro=self.ouro, **kwargs)

    def create(
        self,
        content: Optional["Content"],
        name: str,
        content_markdown: Optional[str] = None,
        content_path: Optional[str] = None,
        description: Optional[Union[str, "Content"]] = None,
        visibility: Optional[str] = None,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
        **kwargs,
    ) -> Post:
        """Create a new Post.

        Provide one of content, content_markdown, or content_path.
        When using content_markdown or content_path, pass content=None.
        """
        content = self._resolve_content(
            content=content,
            content_markdown=content_markdown,
            content_path=content_path,
        )

        post = _strip_none(
            {
                "name": name,
                "description": _coerce_description(description),
                "visibility": visibility,
                "monetization": monetization,
                "price": price,
                **kwargs,
                "source": "api",
                "asset_type": "post",
            }
        )

        request = self.client.post(
            "/posts/create",
            json={
                "post": post,
                "content": content.to_dict(),
            },
        )
        return Post(**self._handle_response(request))

    def retrieve(self, id: str) -> Post:
        """Retrieve a Post by its id."""
        request = self.client.get(f"/posts/{id}")
        return Post(**self._handle_response(request))

    def update(
        self,
        id: str,
        content: Optional["Content"] = None,
        name: Optional[str] = None,
        description: Optional[Union[str, "Content"]] = None,
        visibility: Optional[str] = None,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
        **kwargs,
    ) -> Post:
        """Update a Post by its id."""
        post = _strip_none(
            {
                "name": name,
                "description": _coerce_description(description),
                "visibility": visibility,
                "monetization": monetization,
                "price": price,
                **kwargs,
            }
        )

        request = self.client.put(
            f"/posts/{id}",
            json={
                "post": post,
                "content": content.to_dict() if content is not None else None,
            },
        )
        return Post(**self._handle_response(request))

    def delete(self, id: str) -> None:
        """Delete a Post by its id."""
        request = self.client.delete(f"/posts/{id}")
        self._handle_response(request, raw=True)



from datetime import datetime
from typing import Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel
from typing_extensions import TypedDict


class UserProfile(BaseModel):
    user_id: UUID
    username: Optional[str] = None
    avatar_path: Optional[str] = None
    bio: Optional[str] = None
    is_agent: bool = False


class OrganizationProfile(BaseModel):
    id: UUID
    name: str
    avatar_path: Optional[str] = None
    mission: Optional[str] = None


class DescriptionDict(TypedDict, total=False):
    """Shape of a structured description as returned by the API."""
    json: dict
    text: str


class Asset(BaseModel):
    id: UUID
    user_id: UUID
    user: Optional[UserProfile] = None
    org_id: UUID
    team_id: UUID
    parent_id: Optional[UUID] = None
    organization: Optional[OrganizationProfile] = None
    visibility: str
    asset_type: str
    created_at: datetime
    last_updated: datetime
    name: Optional[str] = None
    description: Optional[Union[str, DescriptionDict]] = None
    metadata: Optional[dict] = None
    monetization: Optional[str] = None
    price: Optional[float] = None
    preview: Optional[dict] = None
    cost_accounting: Optional[str] = None
    cost_unit: Optional[str] = None
    unit_cost: Optional[float] = None
    state: Literal["queued", "in-progress", "success", "error"] = "success"
    source: Literal["web", "api"] = "web"



from __future__ import annotations

import logging
from typing import Dict, List

from ouro._resource import SyncAPIResource
from ouro.models import Route, Service
from ouro.resources.routes import Routes

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Services"]


class Services(SyncAPIResource):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.routes = Routes(*args, **kwargs)

    def retrieve(self, id: str) -> Service:
        """Retrieve a Service by its ID."""
        request = self.client.get(f"/services/{id}")
        return Service(**self._handle_response(request), _ouro=self.ouro)

    def list(self) -> List[Service]:
        """List all services in the current context."""
        request = self.client.get("/services")
        return [Service(**s, _ouro=self.ouro) for s in self._handle_response(request)]

    def read_spec(self, id: str) -> Dict:
        """Get the OpenAPI specification for a service."""
        request = self.client.get(f"/services/{id}/spec")
        return self._handle_response(request)

    def read_routes(self, id: str) -> List[Route]:
        """Get all routes for a service."""
        request = self.client.get(f"/services/{id}/routes")
        return [Route(**r, _ouro=self.ouro) for r in self._handle_response(request)]



from typing import TYPE_CHECKING, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field

from .asset import Asset

if TYPE_CHECKING:
    from ouro import Ouro


class FileData(BaseModel):
    url: str


class FileMetadata(BaseModel):
    name: str
    path: str
    size: int
    type: str
    bucket: Literal["public-files", "files"]
    id: Optional[UUID] = None
    fullPath: Optional[str] = None


class InProgressFileMetadata(BaseModel):
    type: str


class File(Asset):
    metadata: Optional[Union[FileMetadata, InProgressFileMetadata]] = Field(
        default=None,
        union_mode="left_to_right",
    )
    data: Optional[FileData] = None
    _ouro: Optional["Ouro"] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ouro = kwargs.get("_ouro")

    def _require_client(self) -> "Ouro":
        if not self._ouro:
            raise RuntimeError("File object not connected to Ouro client")
        return self._ouro

    def share(
        self,
        user_id: Union[UUID, str],
        role: Literal["read", "write", "admin"] = "read",
    ) -> None:
        """Share this file with another user.

        Args:
            user_id: The UUID of the user to share with.
            role: The role to grant the user (read, write, admin).
        """
        ouro = self._require_client()
        ouro.files.share(str(self.id), user_id, role)

    def read_data(self) -> FileData:
        """Get the file data (signed URL).

        Returns:
            FileData with url property for this file.
        """
        ouro = self._require_client()
        from ouro._resource import SyncAPIResource

        resource = SyncAPIResource(ouro)
        result = resource._handle_response(ouro.client.get(f"/files/{self.id}/data"))

        self.data = FileData(**result)
        return self.data



from __future__ import annotations

import logging
from typing import List, Optional

from ouro._resource import SyncAPIResource, _strip_none
from ouro.models import Conversation

from .content import Content

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Conversations", "Messages"]


class Messages(SyncAPIResource):
    def create(self, conversation_id: str, **kwargs) -> dict:
        json = kwargs.get("json")
        text = kwargs.get("text")
        user_id = kwargs.get("user_id")
        message = _strip_none({
            "json": json,
            "text": text,
            "user_id": user_id,
            **kwargs,
        })

        request = self.client.post(
            f"/conversations/{conversation_id}/messages/create",
            json={"message": message},
        )
        return self._handle_response(request)

    def list(self, conversation_id: str, **kwargs) -> List[dict]:
        request = self.client.get(
            f"/conversations/{conversation_id}/messages", params=kwargs
        )
        return self._handle_response(request)


class ConversationMessages:
    def __init__(self, conversation: "Conversation"):
        self.conversation = conversation
        self._ouro = conversation._ouro

    def create(self, **kwargs) -> dict:
        return Messages(self._ouro).create(self.conversation.id, **kwargs)

    def list(self, **kwargs) -> List[dict]:
        return Messages(self._ouro).list(self.conversation.id, **kwargs)


class Conversations(SyncAPIResource):
    def create(
        self,
        member_user_ids: List[str],
        name: Optional[str] = None,
        summary: Optional[str] = None,
        org_id: Optional[str] = None,
        team_id: Optional[str] = None,
        **kwargs,
    ) -> Conversation:
        """Create a conversation with the specified member user IDs."""
        conversation = _strip_none({
            "name": name,
            "summary": summary,
            "org_id": org_id,
            "team_id": team_id,
            "metadata": {"members": member_user_ids},
            **kwargs,
        })

        request = self.client.post(
            "/conversations/create",
            json={"conversation": conversation},
        )
        return Conversation(**self._handle_response(request), _ouro=self.ouro)

    def retrieve(self, conversation_id: str) -> Conversation:
        """Retrieve a conversation by id."""
        request = self.client.get(f"/conversations/{conversation_id}")
        return Conversation(**self._handle_response(request), _ouro=self.ouro)

    def list(
        self,
        org_id: Optional[str] = None,
        limit: int = 20,
        offset: int = 0,
    ) -> List[Conversation]:
        """List conversations with optional org filter and pagination."""
        params = {
            "limit": limit,
            "offset": offset,
        }
        if org_id is not None:
            params["org_id"] = org_id

        request = self.client.get("/conversations", params=params)
        return [
            Conversation(**c, _ouro=self.ouro)
            for c in self._handle_response(request)
        ]

    def update(self, conversation_id: str, **kwargs) -> Conversation:
        """Update a conversation."""
        request = self.client.put(
            f"/conversations/{conversation_id}", json={"conversation": kwargs}
        )
        return Conversation(**self._handle_response(request), _ouro=self.ouro)



from typing import TYPE_CHECKING, Dict, List, Optional

from pydantic import BaseModel

from .asset import Asset
from .route import Route

if TYPE_CHECKING:
    from ouro import Ouro


class ServiceMetadata(BaseModel):
    base_url: str
    spec_path: str
    authentication: str


class Service(Asset):
    metadata: Optional[ServiceMetadata] = None
    _ouro: Optional["Ouro"] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ouro = kwargs.get("_ouro")

    def _require_client(self) -> "Ouro":
        if not self._ouro:
            raise RuntimeError("Service object not connected to Ouro client")
        return self._ouro

    def read_spec(self) -> Dict:
        """Get the OpenAPI specification for this service."""
        ouro = self._require_client()
        return ouro.services.read_spec(str(self.id))

    def read_routes(self) -> List[Route]:
        """Get all routes for this service."""
        ouro = self._require_client()
        return ouro.services.read_routes(str(self.id))

    def use_route(self, route_name_or_id: str, **kwargs) -> Dict:
        """Use/execute a specific route of this service."""
        ouro = self._require_client()
        return ouro.services.routes.use(f"{self.id}/{route_name_or_id}", **kwargs)



import logging
import time
from typing import Callable, Optional

import socketio

log = logging.getLogger(__name__)


class OuroWebSocket:
    def __init__(self, ouro):
        self.ouro = ouro
        self.sio = socketio.Client(
            reconnection=True,
            reconnection_attempts=5,
            reconnection_delay=1,
        )
        self.setup_event_handlers()

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    def setup_event_handlers(self):
        @self.sio.event
        def connect():
            log.info("Connected to websocket")

        @self.sio.event
        def disconnect():
            log.warning("Disconnected from websocket")

        @self.sio.event
        def connect_error(data):
            log.error(f"Connection error: {data}")

    def connect(self, access_token: Optional[str] = None) -> None:
        try:
            self.sio.connect(
                self.ouro.websocket_url,
                retry=True,
                namespaces=["/"],
                auth={
                    "access_token": access_token or self.ouro.access_token,
                },
            )
            self.sio.sleep(1)
        except Exception as e:
            log.error(f"Failed to connect to websocket: {e}")

    def disconnect(self):
        self.sio.disconnect()

    def refresh_connection(self, access_token: Optional[str] = None):
        self.disconnect()
        self.sio.sleep(1)
        self.connect(access_token)

    def handle_disconnect(self):
        max_retries = 5
        retry_delay = 1  # Start with 1 second delay

        for attempt in range(max_retries):
            log.info(f"Attempting to reconnect (attempt {attempt + 1}/{max_retries})")
            try:
                self.connect()
                if self.is_connected:
                    log.info("Reconnection successful")
                    return
            except Exception as e:
                log.error(f"Reconnection attempt failed: {e}")

            # Exponential backoff
            retry_delay *= 2
            log.info(f"Waiting {retry_delay} seconds before next attempt")
            time.sleep(retry_delay)

        log.error("Failed to reconnect after maximum attempts")

    def on(self, event: str, handler: Callable):
        self.sio.on(event, handler)

    def emit(self, event, data):
        if not self.is_connected:
            log.warning("Attempted to emit event while disconnected. Reconnecting...")
            self.connect()
        return self.sio.emit(event, data)

    def __del__(self):
        if self.is_connected:
            self.disconnect()



from __future__ import annotations

import logging
import mimetypes
import os
import uuid
from typing import Literal, Optional, Union

from ouro._resource import SyncAPIResource, _coerce_description, _strip_none
from ouro.models import File

from .content import Content

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Files"]


def _build_file_metadata(
    file_id: str,
    file_name: str,
    bucket: str,
    path_on_storage: str,
    mime_type: str | None,
    server_metadata: dict,
) -> dict:
    """Merge local upload info with server-extracted metadata."""
    return {
        "id": file_id,
        "name": file_name,
        "bucket": bucket,
        "path": path_on_storage,
        "type": mime_type,
        "mimeType": mime_type,
        **server_metadata,
    }


class Files(SyncAPIResource):
    def create(
        self,
        name: str,
        visibility: str,
        file_path: Optional[str] = None,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
        description: Optional[Union[str, "Content"]] = None,
        **kwargs,
    ) -> File:
        """Create a File."""
        log.debug("Creating a file")
        if not file_path:
            log.warning("No file path provided, creating a file stub. Update it later.")
            file = {
                "id": str(uuid.uuid4()),
                "name": name,
                "visibility": visibility,
                "monetization": monetization,
                "price": price,
                "description": _coerce_description(description),
                **kwargs,
                "asset_type": "file",
                "state": "in-progress",
                "source": "api",
            }
        else:
            id = str(uuid.uuid4())
            with open(file_path, "rb") as f:
                mime_type = mimetypes.guess_type(file_path)[0]
                file_extension = os.path.splitext(file_path)[1]

                bucket = "public-files" if visibility == "public" else "files"
                bucket_folder = f"{self.ouro.user.id}"
                file_name = f"{id}{file_extension}"
                path_on_storage = f"{bucket_folder}/{file_name}"

                log.info(f"Uploading file to {path_on_storage} in the {bucket} bucket")
                self.ouro.supabase.storage.from_(bucket).upload(
                    file=f,
                    path=path_on_storage,
                    file_options={"content-type": mime_type},
                )

                response = self.ouro.supabase.storage.from_(bucket).list(
                    bucket_folder,
                    {
                        "limit": 1,
                        "offset": 0,
                        "sortBy": {"column": "name", "order": "desc"},
                        "search": id,
                    },
                )
                file = response[0]
                if file["name"] != file_name:
                    raise RuntimeError(
                        f"Upload verification failed: expected '{file_name}', "
                        f"got '{file['name']}'"
                    )

            file_id = file["id"]
            meta_data = self._handle_response(
                self.client.get(f"/files/{file_id}/metadata")
            )

            metadata = _build_file_metadata(
                file_id, file_name, bucket, path_on_storage,
                mime_type, meta_data["metadata"],
            )

            file = {
                "id": file_id,
                "name": name,
                "visibility": visibility,
                "monetization": monetization,
                "price": price,
                "description": _coerce_description(description),
                **kwargs,
                "source": "api",
                "metadata": metadata,
                "preview": meta_data["preview"],
                "asset_type": "file",
            }

        file = _strip_none(file)

        request = self.client.post("/files/create", json={"file": file})
        data = self._handle_response(request)
        return File(**data, _ouro=self.ouro)

    def retrieve(self, id: str) -> File:
        """Retrieve a File by its ID."""
        data = self._handle_response(self.client.get(f"/files/{id}"))

        try:
            file_data = self._handle_response(self.client.get(f"/files/{id}/data"))
        except Exception:
            file_data = None

        data["data"] = file_data
        return File(**data, _ouro=self.ouro)

    def update(
        self,
        id: str,
        file_path: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[Union[str, "Content"]] = None,
        visibility: Optional[str] = None,
        **kwargs,
    ) -> File:
        """Update a file by ID.

        Pass file_path to replace the file data with a new file from the local filesystem.
        Pass name, description, or visibility to update metadata.
        """
        log.debug("Updating a file")

        update_params = _strip_none({
            "name": name,
            "description": _coerce_description(description),
            "visibility": visibility,
        })
        update_params.update(kwargs)

        file: dict = {"id": str(id), **update_params}
        existing = self.retrieve(id)

        if file_path:
            with open(file_path, "rb") as f:
                mime_type = mimetypes.guess_type(file_path)[0]
                file_extension = os.path.splitext(file_path)[1]

                bucket = "public-files" if existing.visibility == "public" else "files"
                path_on_storage = f"{self.ouro.user.id}/{id}{file_extension}"

                log.info(f"Uploading file to {path_on_storage} in the {bucket} bucket")
                request = self.ouro.supabase.storage.from_(bucket).upload(
                    file=f,
                    path=path_on_storage,
                    file_options={"content-type": mime_type},
                )
                file_data = request.json()

            file_id = file_data["Id"]
            meta_info = self._handle_response(
                self.client.get(f"/files/{file_id}/metadata")
            )

            metadata = _build_file_metadata(
                file_id, f"{id}{file_extension}", bucket, path_on_storage,
                mime_type, meta_info["metadata"],
            )

            file = {
                **file,
                "metadata": metadata,
                "preview": meta_info["preview"],
                "asset_type": "file",
            }

        file = _strip_none(file)

        request = self.client.put(f"/files/{id}", json={"file": file})
        data = self._handle_response(request)
        return File(**data, data=None, _ouro=self.ouro)

    def delete(self, id: str) -> None:
        """Delete a file."""
        request = self.client.delete(f"/files/{id}")
        self._handle_response(request)

    def share(
        self,
        file_id: str,
        user_id: Union[uuid.UUID, str],
        role: Literal["read", "write", "admin"] = "read",
    ) -> None:
        """Share a file with another user."""
        request = self.client.put(
            f"/elements/common/{file_id}/share",
            json={"permission": {"user": {"user_id": str(user_id)}, "role": role}},
        )
        self._handle_response(request)
